#define _POSIX_SOURCE

#include <lauxlib.h>
#include <stdlib.h>
#include <linux/limits.h>

#define SEED 42
#define OBSTACLE -1
#define MOVE_N 4

typedef struct {
    int w, h;
    unsigned int seed;
    char *a, *b;
} model_t;

#define A(i, j) M->a[(M->w + 2) * (i + 1) + j + 1]
#define B(i, j) M->b[(M->w + 2) * (i + 1) + j + 1]

static int rand_2(unsigned int *seed) {
    return rand_r(seed) & 1;
}

static double drand_01(unsigned int *seed) {
    return (double) rand_r(seed) / INT_MAX;
}

model_t *new(lua_State *L) {
    model_t *M = malloc(sizeof(model_t));

    M->seed = SEED;

    lua_getglobal(L, "height"); M->h = luaL_checkint(L, -1);
    lua_getglobal(L, "width");  M->w = luaL_checkint(L, -1);
    lua_pop(L, 2);

    int bytes = (M->w + 2) * (M->h + 2);
    M->a = calloc(bytes, 1);
    M->b = calloc(bytes, 1);

    lua_getglobal(L, "locations");

    lua_len(L, -1);
    int n = luaL_checkint(L, -1);
    lua_pop(L, 1);

    for (int l = 1; l <= n ; l++) {
        lua_rawgeti(L, -1, l);

        lua_getfield(L, -1, "density");
        double p = lua_tonumber(L, -1) / MOVE_N;

        lua_getfield(L, -2, "area");
        int r[4];
        for (int i = 1; i <= 4; i++) {
            lua_rawgeti(L, -1, i);
            r[i - 1] = luaL_checkint(L, -1);
            lua_pop(L, 1);
        }
        lua_pop(L, 3);

        for (int i = r[0]; i < r[2]; i++)
        for (int j = r[1]; j < r[3]; j++) {
            A(i, j) = 0;
            char b = 1;
            for (int k = 0; k < MOVE_N; k++) {
                if (drand_01(&M->seed) < p)
                    A(i, j) |= b;
                b <<= 1;
            }
        }
    }

    for (int i = -1; i < M->h + 1; i++) A(i, -1) = B(i, -1) = A(i, M->w) = B(i, M->w) = OBSTACLE;
    for (int j = -1; j < M->w + 1; j++) A(-1, j) = B(-1, j) = A(M->h, j) = B(M->h, j) = OBSTACLE;

    return M;
}

void del(model_t *M) {
    free(M->a);
    free(M->b);
    free(M);
}

static void collide(model_t *M) {
    char state[] = {5, 10};
#pragma omp parallel for
    for (int i = 0; i < M->h; i++)
    for (int j = 0; j < M->w; j++)
        if (A(i, j) == 5 || A(i, j) == 10)
            A(i, j) = state[rand_2(&M->seed)];
}

static void stream(model_t *M) {
#define ML(a, b, c) B(i, j) |= A(a, b) != OBSTACLE ? A(a, b) & c : (A(i, j) & (c << 2)) >> 2
#define MR(a, b, c) B(i, j) |= A(a, b) != OBSTACLE ? A(a, b) & c : (A(i, j) & (c >> 2)) << 2
#pragma omp parallel for
    for (int i = 0; i < M->h; i++)
    for (int j = 0; j < M->w; j++) {
        B(i, j) = 0;

        ML(i + 1, j, 1);
        ML(i, j - 1, 2);

        MR(i - 1, j, 4);
        MR(i, j + 1, 8);
    }
}

void step(model_t *M, int iter) {
    collide(M);
    stream(M);
    char *t = M->a; M->a = M->b; M->b = t;
}

void dump(model_t *M, int iter) {
    char path[PATH_MAX];
    snprintf(path, PATH_MAX, "%08d.pgm", iter);
    FILE *fp = fopen(path, "w");
    fprintf(fp, "P5\n%d\n%d\n%d\n", M->w, M->h, 255);
    char pgm[M->h][M->w];
    for (int i = 0; i < M->h; i++)
        for (int j = 0; j < M->w; j++)
            pgm[i][j] = 255.0 / MOVE_N * __builtin_popcount(A(i, j));
    fwrite(pgm, 1, M->h * M->w, fp);
    fclose(fp);
}
