#define _POSIX_SOURCE

#include <lauxlib.h>
#include <stdlib.h>
#include <linux/limits.h>
#include <stdint.h>
#include <math.h>

#include "util.h"
#include "vector.h"
#include "map.h"

#define SEED 42
#define SIZE_N 2

typedef int32_t cell_t;
typedef int point_t[SIZE_N];

typedef vector_t(cell_t) class_t;
typedef vector_t(class_t) rules_t;
typedef vector_t(point_t) poly_t;

typedef vector_t(int) vint_t;
typedef map_t(int, vint_t) mvint_t;

typedef struct {
    point_t size;
    poly_t velo;
    unsigned int seed;
    rules_t rules;
    cell_t *a;
    cell_t *b;
} model_t;

#define A(i, j) M->a[M->size[1] * ((i + M->size[0]) % M->size[0]) + (j + M->size[1]) % M->size[1]]
#define B(i, j) M->b[M->size[1] * ((i + M->size[0]) % M->size[0]) + (j + M->size[1]) % M->size[1]]

static int rand_n(unsigned int *seed, int n) {
    return rand_r(seed) % n;
}

static double drand_01(unsigned int *seed) {
    return (double) rand_r(seed) / INT_MAX;
}

static void init_size(lua_State *L, model_t *M) {
    lua_getglobal(L, "height");
    M->size[0] = luaL_checkint(L, -1);

    lua_getglobal(L, "width");
    M->size[1] = luaL_checkint(L, -1);

    lua_pop(L, 2);
}

static void read_poly(lua_State *L, poly_t *v) {
    lua_len(L, -1);
    int size = luaL_checkint(L, -1);
    lua_pop(L, 1);

    vector_alloc((*v), size); v->size = size;

    for (int i = 0; i < size; i++) {
        lua_rawgeti(L, -1, i + 1);

        for (int j = 1; j <= SIZE_N; j++) {
            lua_rawgeti(L, -j, j);
            v->_[i][j - 1] = luaL_checkint(L, -1);
        }

        lua_pop(L, 1 + SIZE_N);
    }
}

static void init_velo(lua_State *L, model_t *M) {
    lua_getglobal(L, "particle_velocities");
    read_poly(L, &M->velo);
    lua_pop(L, 1);
}

static void fill_cell(model_t *M, double dens, point_t c) {
    double prob = dens / M->velo.size;

    A(c[0], c[1]) = 0;

    cell_t p = 1;
    for (int k = 0; k < M->velo.size; k++) {
        if (drand_01(&M->seed) < prob)
            A(c[0], c[1]) |= p;
        p <<= 1;
    }
}

static void _add_line(mvint_t *is, point_t a, point_t b, int s) {
    if (a[1] > b[1]) swap(a, b);

    int di = abs(b[0] - a[0]);
    int dj = b[1] - a[1];
    int e = dj / 2;
    int z = a[0] < b[0] ? 1 : -1;

    int i = a[0];
    for (int j = a[1]; j <= b[1]; j++) {
        if (s) {
            vint_t js = {}; map_get((*is), j, js);
            int f = 1;
            for (int k = 0; k < js.size; k++)
                if (abs(js._[k] - i) <= 2) {
                    f = 0;
                    break;
                }
            if (f) {
                vector_push(js, i);
                map_set((*is), j, js);
            }
        }

        e -= di;
        if (e < 0) {
            if (!s) {
                vint_t js = {}; map_get((*is), i, js);
                int f = 1;
                for (int k = 0; k < js.size; k++)
                    if (abs(js._[k] - j) <= 2) {
                        f = 0;
                        break;
                    }
                if (f) {
                    vector_push(js, j);
                    map_set((*is), i, js);
                }
            }

            i += z;
            e += dj;
        }
    }
}

static void add_line(mvint_t *is, point_t a, point_t b) {
    int s = abs(a[0] - b[0]) > abs(a[1] - b[1]);
    point_t c, d;
    for (int i = 0; i < SIZE_N; i++) {
        c[i] = a[s ? SIZE_N - 1 - i : i];
        d[i] = b[s ? SIZE_N - 1 - i : i];
    }
    _add_line(is, c, d, s);
}

static int cmp(const void *a, const void *b) {
    return ((*(int *)a) - (*(int *)b));
}

static void fill_poly(model_t *M, double dens, poly_t poly) {
    mvint_t is = {};

    for (int i = 0; i < poly.size; i++) {
        int j = (i + 1) % poly.size;
        if (poly._[i][0] != poly._[j][0])
            add_line(&is, poly._[i], poly._[j]);
    }

    for (int k = 0; k < is.size; k++) {
        int i = is._[k].k;
        vint_t js = is._[k].v;
        if (js.size & 1) continue;
        qsort(js._, js.size, sizeof(int), cmp);
        for (int t = 0; t < js.size; t += 2) {
            int jf = js._[t];
            int jl = js._[t + 1];
            for (int j = jf; j <= jl; j++)
                fill_cell(M, dens, (point_t) {i, j});
        }
    }

    for (int i = 0; i < is.size; i++) {
        printf("%d:", is._[i].k);
        for (int j = 0; j < is._[i].v.size; j++)
            printf(" %d", is._[i].v._[j]);
        printf("\n");
    }
}

static void init_grid(lua_State *L, model_t *M) {
    int bytes = M->size[0] * M->size[1] * sizeof(cell_t);
    M->a = calloc(bytes, 1);
    M->b = calloc(bytes, 1);

    lua_getglobal(L, "locations");

    lua_len(L, -1);
    int size = luaL_checkint(L, -1);

    for (int i = 1; i <= size; i++) {
        lua_rawgeti(L, -2, i);

        lua_getfield(L, -1, "density");
        double dens = lua_tonumber(L, -1);

        lua_getfield(L, -2, "area");
        poly_t poly = {}; read_poly(L, &poly);

        fill_poly(M, dens, poly);

        lua_pop(L, 1 + 2);
    }

}

static int eq(model_t *M, cell_t a, cell_t b) {
    int ma = 0;
    int mb = 0;
    point_t va = {};
    point_t vb = {};

    cell_t p = 1;
    for (int i = 0; i < M->velo.size; i++) {
        if (p & a) {
            for (int j = 0; j < SIZE_N; j++)
                va[j] += M->velo._[i][j];
            ma++;
        }
        if (p & b) {
            for (int j = 0; j < SIZE_N; j++)
                vb[j] += M->velo._[i][j];
            mb++;
        }
        p <<= 1;
    }

    int is_eq = ma == mb;
    for (int i = 0; i < SIZE_N; i++)
        is_eq = is_eq && va[i] == vb[i];

    return is_eq;
}

static void generate_rules(model_t *M) {
    const int state_n = 1 << M->velo.size;
    vector_alloc(M->rules, state_n);
    for (cell_t i = 0; i < state_n; i++) {
        class_t class = {};
        for (cell_t j = 0; j < state_n; j++)
            if (eq(M, i, j)) vector_push(class, j);
        vector_push(M->rules, class);
    }
}

#define ACR "\x1b[31m"
#define ACG "\x1b[32m"
#define ACY "\x1b[33m"
#define ACB "\x1b[34m"
#define ACM "\x1b[35m"
#define ACC "\x1b[36m"
#define ACE "\x1b[0m"

static void print_rules(model_t *M) {
    for (cell_t i = 0; i < M->rules.size; i++) {
        printf("%d:", i);
        for (cell_t j = 0; j < M->rules._[i].size; j++)
            printf(" %d", M->rules._[i]._[j]);
        printf("\n");
    }
}

model_t *new(lua_State *L) {
    model_t *M = calloc(1, sizeof(model_t));

    M->seed = SEED;

    init_size(L, M);
    init_velo(L, M);
    init_grid(L, M);

    generate_rules(M);
//    print_rules(M);

    return M;
}

void del(model_t *M) {
    free(M->a);
    free(M->b);
    free(M->velo._);

    for (cell_t i = 0; i < M->rules.size; i++)
        free(M->rules._[i]._);
    free(M->rules._);

    free(M);
}

static void collide(model_t *M) {
    for (int i = 0; i < M->size[0]; i++)
        for (int j = 0; j < M->size[1]; j++) {
            class_t class = M->rules._[A(i, j)];
            A(i, j) = class._[rand_n(&M->seed, class.size)];
        }
}

static void stream(model_t *M) {
    for (int i = 0; i < M->size[0]; i++)
        for (int j = 0; j < M->size[1]; j++) {
            B(i, j) = 0;
            cell_t p = 1;
            for (int k = 0; k < M->velo.size; k++) {
                B(i, j) |= A(i - M->velo._[k][0], j - M->velo._[k][1]) & p;
                p <<= 1;
            }
        }
}

void step(model_t *M, int iter) {
    collide(M);
    stream(M);
    swap(M->a, M->b);
}

void dump(model_t *M, int iter) {
    char path[PATH_MAX];
    snprintf(path, PATH_MAX, "%08d.pgm", iter);
    FILE *fp = fopen(path, "w");

    fprintf(fp, "P5\n%d\n%d\n%d\n", M->size[1], M->size[0], 255);

    char pgm[M->size[0]][M->size[1]];
    for (int i = 0; i < M->size[0]; i++)
        for (int j = 0; j < M->size[1]; j++)
            pgm[i][j] = 255.0 / M->velo.size * __builtin_popcount(A(i, j));

    fwrite(pgm, 1, M->size[0] * M->size[1], fp);

    fclose(fp);
}
